'''
:File: __init__.py
:Author: Jayesh Joshi
:Email: jayeshjo1@utexas.edu

Conplyent provides a remote shell executor built completely on python. This is
built to provide the ability to test applications remotely and verify system
stability/etc.
'''


# Console Executor Direct link
from .console import ConsoleExecutor

# Exceptions
from .exceptions import ConsoleExecTimeout, ClientTimeout, ZMQPairTimeout

# Main Client/Server
from . import client, server

# Server globals and register commands
from .server import MSG_PORT, INVALID_PARAMETER, SUCCESS, ERROR, register_command, register_background_command



'''
:File: client.py
:Author: Jayesh Joshi
:Email: jayeshjo1@utexas.edu
'''

import time
import inspect
import os
from threading import local

from ._zmq_pair import ZMQPair
from ._msg import MSG, MSGType
from .exceptions import ZMQPairTimeout, ClientTimeout
from ._decorators import timeout
from ._general import logger
from .server import INVALID_PARAMETER, ERROR

_pool = local()


def add(hostname, port=9922):
    '''
    Clients can be added using this method which registers the a connection to a
    local pool to allow for creation of many clients. This creates a class
    :class:`client.ClientConnection` which will allow users to communicate
    with the server existing on a different process (local or remote).

    This does not establish the connection.

    :param hostname: Host name of the server.
    :type hostname: str
    :param port: Port number that the server is listening to.
    :type port: int
    '''
    connection = ClientConnection(_new_connection(ZMQPair(dest_ip=hostname, port=port)))
    return connection


class ClientConnection():
    '''
    Client connection class that provides the means to connect to a local or
    remote server and execute commands. These commands are defined in the
    server module. This class will request to sync methods with the
    remote server which will provide all available methods. Users using the
    ClientConnection class can easily send commands over to the remote server as
    if just calling class methods.

    :Example:

    >>> from conplyent import client
    >>> my_conn = client.add_client("localhost")
    >>> my_conn.connect()
    >>> my_conn.ls()

    These methods are not initially defined in the client class so the
    Connection will not be able to provide the methods intiially. Attempting to
    run commands at this point will result in AttributeError.

    Once the connection has been established, users can use
    :meth:`client.ClientConnection.server_methods` to view all of the commands.
    This method does not provide any keyword parameters though. May look into
    doing so in the future. When using these commands, users can access the
    server methods using any args or kwargs combination as defined in the server
    methods. These methods will have the same requirements as defined in the
    server. I.e. order sent to server is the order defined in the server
    methods. And the kwargs follow the same name:value format as the server.

    Both of these are valid:

    >>> my_conn.cd("..")
    >>> my_conn.cd(dest="..")

    Any errors resulted by user input will be passed back to the client. These
    will not crash the server. For example, if we pass in an int to 'cd'
    command, this command will output the complete traceback provided by the
    server.

    :cd(1):

    >>> TypeError('chdir: path should be string, bytes or os.PathLike, not int')

    When using these server methods, users can provide three additional kwargs
    to the method that are processed client side. These are "timeout",
    "complete", "echo_response", "max_interval", and "raise_error".

    Timeout defines the amount of time that we want to wait for the send command
    to go through. This could come into play for scenarios where you have a
    bunch of network activity on the host so that send command may take a while
    to go through ZMQ. By default timeout is None, so the send will be
    guaranteed to go through (though no promises on how long it may take). On a
    typical system where you have a more relaxed network activity, this send
    should go through immedaitely.

    Complete can be disabled to have more control over how the this class
    receives outputs from the server. If complete is set to True (as is
    default), once the command has been sent, this class will continually
    attempt to receives the output from the server and append to a list that is
    returned. If complete is set to False, these methods will return a
    :class:`client.ClientConnection.ClientListener` instance which the user can
    use themselves to listen to the output in real time. This is mostly useful
    when we are running background tasks. These tasks will still continue to be
    executed in the server but we don't need to wait for it. Examples of use
    cases: If we need to send input to the task, if we want to kill it
    prematurely, or if we want to parse and process the output in real time.
    See description of jobs for more information.

    :complete=False:

    >>> listener = my_conn.exec("some_long_command", complete=False)
    >>> ... do something else ...
    >>> for response in iter(listener.next, None):
    >>>     ... do something with response ...

    Echo Response will simply print to stdout the response in real time. This
    can be useful when users have enabled complete.

    Max Interval determines the amount of time that the listener should wait
    between each message. This should be used when you expect that the program
    being run on the system under test should respond every x interval and not
    doing so is a significance of the server crashing, whether due to
    application error or due to server in an unstable state. If the server does
    not respond between each messages larger than user set max_interval, then
    listener will throw a ClientTimeout error. Max Interval is defined in
    seconds, and can be sent as a floating number to specify lower granularity.
    For cases where complete is set to False, max_interval can later be adjusted
    through the property exposed by ClientListener.

    :max_interval=#:

    >>> my_conn.exec("check_system_stability", max_interval=10, complete=True)

    Raise Error is by default set to False. Command sent to the server can
    return either INVALID_PARAMETER, which is a field returned by server
    commands noting that the parameter sent by user is invalid, or ERROR, noting
    that the command failed due to some exception. If Raise Error is set to
    True, then INVALID_PARAMETER responses will raise a ValueError, and ERROR
    responses will raise the error retrieved.
    '''

    def __init__(self, conn_id, enter_connect=True, enter_connect_timeout=None):
        self.__conn_id = conn_id
        self.__synced = False
        self.__initial_methods = [i[0] for i in [k for k in inspect.getmembers(self, inspect.ismethod)]]
        self.__enter_connect = enter_connect
        self.__enter_connect_timeout = enter_connect_timeout

    def __enter__(self):
        if(self.__enter_connect):
            self.connect(timeout=self.__enter_connect_timeout)
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def connect(self, timeout=None):
        '''
        Finds the ZMQPair from the local pool and try to establish connection.
        Once the connection is established, will wait until we get a response
        from the server guaranteeing that the server is alive. ZMQ connection
        success just means that we have been able to own the port of the
        localhost to be able to send commands, does not guarantee that we have
        made a handshake with the server.

        Users may pass in timeout for how long we should wait to see if the
        server responds. This can be useful in the cases where we have pulsed a
        reboot or shutdown request to the server so this connect method will
        wait for the server to come back alive.

        :Optional:

        :param timeout: Amount of time in seconds to wait for the connection to
            be established. None = wait forever (nonblocking).
        :type timeout: (int)

        :raises ConnectionError: If server does not respond in the desired
            amount of time.
        '''
        connection = _get_connection(self.__conn_id)
        connection.connect()
        if(timeout):
            logger.info("Waiting {}s for {} to respond...".format(timeout, self.__conn_id))
        else:
            logger.info("Waiting for {} to respond...".format(self.__conn_id))
        if(not(connection.pulse(timeout))):
            raise ConnectionError("Server not responding at {}".format(self.__conn_id))
        if(not(self.__synced)):
            self.__sync_attr()
            self.__synced = True
            logger.info("Synced with server.")

    def close(self):
        '''
        Completely closes the socket used to communicate to the server. This
        should only be called when the client is finished communicating with the
        server. For temporary disconnects, use
        :meth:`client.ClientConnection.disconnect`.

        This will prevent this connection class from working in the future. User
        will need to add a new connection to communicate with the server.
        '''
        connection = _get_connection(self.__conn_id)
        connection.close()
        logger.info("Closing connection to {}".format(self.__conn_id))

    def disconnect(self):
        '''
        Disconnects with the server. Commands can no longer be sent.

        This is useful for cases when the server drops connection and we need to
        re-establish connection by disconnecting the socket prematurely.
        '''
        connection = _get_connection(self.__conn_id)
        connection.disconnect()
        logger.info("Disconnecting {}".format(self.__conn_id))

    def server_methods(self):
        '''
        When user calls connect, this connection class will sync with the server
        to gather all of the possible commands that can be sent to the server.
        These commands are added as new attributes for this class.

        .. note:: This should only be called once the connection has been
            established. Otherwise will return nothing.
        '''
        commands = [i[0] for i in [k for k in inspect.getmembers(self, inspect.ismethod)]]
        return list(filter(lambda k: not (k[0] == "_" or k in self.__initial_methods), commands)) + ["transfer"]

    def heartbeat(self, timeout=None):
        '''
        Checks to see if the server is still alive by sending a simple Heartbeat
        request. Even if server is executing tasks in the background, the main
        server is mainly just waiting for client requests. This means that This
        request should be honored immediately. By default, this method will wait
        forever for a response from the server. Users may pass in a timeout to
        see if the server will respond in that amount of time.

        :Optional:

        :param timeout: Time in seconds to wait for a heartbeat response
        :type timeout: int

        :returns: True if the server responded. False if timed out.
        '''
        connection = _get_connection(self.__conn_id)
        return connection.pulse(timeout=timeout)

    def transfer(self, src, dest, chunk=1024, timeout=None, **kwargs):
        '''
        Transfers file from client to server.

        :param src: Path to file on client
        :type src: str
        :param dest: Destination to save on server
        :type dest: str

        :Optional:

        :param chunk: Amount in bytes to send per transaction. (1024)
        :type chunk: int
        :param timeout: Time to wait for commands to be sent.
        :type timeout:
        '''
        self.mkdirs(os.path.dirname(dest), exist_ok=True, timeout=timeout, **kwargs)
        append = False
        with open(src, "rb") as file:
            while(True):
                data = file.read(chunk)
                if(data == b''):
                    break
                else:
                    self.wrfile(dest, data, append=append, timeout=timeout, **kwargs)
                    append = True

    def _command_server(self, cmd_id, *args, timeout=None, complete=True, echo_response=False, max_interval=None,
                        raise_error=False, **kwargs):
        listener = self.__send_command(cmd_id, timeout=timeout, max_interval=max_interval, *args, **kwargs)
        if(complete):
            output = list()
            for response in iter(listener.next, None):
                output.append(response)
                if(echo_response):
                    print(response.rstrip() if type(response) is str else response)
            if(raise_error):
                if(listener.exit_code == ERROR):
                    raise output[-1]
                elif(listener.exit_code == INVALID_PARAMETER):
                    raise ValueError(output[-1])
            return output[1:]  # ignore the echo of command
        else:
            return listener

    def __send_command(self, cmd_id, *args, timeout=None, max_interval=None, **kwargs):
        connection = _get_connection(self.__conn_id)
        msg = MSG(MSGType.COMMAND, cmd_id=cmd_id, args=args, keywargs=kwargs)
        connection.send_msg(msg, timeout)
        logger.debug("Sent Message: {}".format(msg))
        return ClientConnection.ConnectionListener(connection, max_interval)

    def __sync_attr(self, timeout=None):
        connection = _get_connection(self.__conn_id)
        connection.send_msg(MSG(MSGType.SYNC, request=True))
        extra_list = list()
        while(True):
            msg = connection.recv_msg(timeout=timeout)
            if(msg.type == MSGType.SYNC):
                for name in msg.args[0]:
                    setattr(ClientConnection, name, _wrap_server_methods(self, name))
                break
            else:
                extra_list.append(msg)
        for extra in extra_list:
            connection.requeue_msg(extra)

    class ConnectionListener():
        '''
        This class is created when users send command over to the server. This
        class is instantiated by :class:`client.ClientConnection` and should not
        be initialized by the user.

        Used to read in message updates by the job and keep track of attributes
        related to the job such as described below. Initially, this class will
        wait for a handshake response by the server acknowledging that the
        server has begun the request. This could wait forver if the server fails
        to acknowlege the request, whether if the server's network is full or if
        the server has crashed. The later is highly unlikely to be the case,
        since this listener object will only be created after the server has
        been checked for a heartbeat.
        '''

        def __init__(self, connection, max_interval):
            self.__connection = connection
            self.__find_ack()
            self._done = False
            self._exit_code = None
            self._max_interval = max_interval
            self.__curr_msg = 0

        def __enter__(self):
            return self

        def __exit__(self, type, value, traceback):
            pass

        @property
        def done(self):
            '''
            Used to determine whether a job is complete or not. This is only set
            when Listener has finished reading all of the messages. If the job
            has completed on the server side, but the user has not gone through
            all of the messages using
            :meth:`client.ClientConnection.ClientListener.next`, then this will
            still read as False.

            :getter: (bool) True if job has been completed. False otherwise.
            '''
            return self._done

        @property
        def exit_code(self):
            '''
            Server methods exit with a return code. Default conplyent server
            methods follow 0 = success, 1 = error in runtime, -1 = bad
            parameter.

            :getter: (int) Exit code of server method. None if job has not
                completed.
            '''
            return self._exit_code

        @property
        def id(self):
            '''
            Unique ID given to each job. This ID is required for some of the
            server methods which interacts with backkground jobs, such as
            sending input or killing the said job.

            :getter: (int) Unique ID assigned to job.
            '''
            return self._id

        @property
        def max_interval(self):
            '''
            Max interval between two messages that this listener will receive.
            If this value is None, there will be no timeout. If this value is
            a positive integer, then Listener will timeout if messages are not
            retrieved within the max_interal.

            :getter: Get the max interval setting of this listener
            :setter: Sets the max interval setting of this listener. Setting
                must be either None, or positive integer
            '''
            return self._max_interval

        @max_interval.setter
        def max_interval(self, value):
            assert value is None or type(value) is int and value >= 0, "Timeout must be None or positive integer"
            self._max_interval = value

        def next(self):
            '''
            Checks to see if there has been any messages for this job. By
            default, will keep waiting until a message has arrived. Users can
            override this passing in timeout to wait only a set amount of time
            for each message to the initialization of this listener.

            :returns: Update provided by the job, or None if: (Job has
                Completed, timed out waiting for message, or connection with
                server has been interrupted)

            :raises ClientTimeout: If user specifies timeout and message is not
                received in the given time.

            .. warning:: If the job has already completed, this method should
                not be called and will throw an AssertionError.
            '''
            self.__response = None
            assert not(self._done), "Server has already completed job..."
            try:
                self.__receive_message(timeout=self._max_interval, exception=ClientTimeout)
            except (ZMQPairTimeout):
                raise ClientTimeout("Server did not respond in {} s".format(timeout))
            return self.__response

        def clear_messages(self):
            '''
            Goes through all of the messages sent by the server in regards to
            this command and clears them out. This will NOT return anything.
            Does not raise ClientTimeout error. Users should check with done
            property to determine whether the command is complete or not.

            Useful if you don't care about the output and want to continue until
            timeout.
            '''
            try:
                for msg in iter(self.next, None):
                    pass
            except ClientTimeout:
                return

        @timeout(name="Listening")
        def __receive_message(self, *args, **kwargs):
            while(True):
                msg = self.__connection.recv_msg(timeout=kwargs["timeout"], msg_id=self.id)
                if((msg.type == MSGType.COMPLETE or msg.type == MSGType.DETAILS) and
                   msg.request_id == self._id and msg.msg_num == self.__curr_msg):
                    self.__curr_msg += 1
                    if(msg.type == MSGType.COMPLETE):
                        self._done = True
                        self.__response = None
                        self._exit_code = msg.exit_code
                    else:
                        self.__response = msg.details
                    break
                else:
                    self.__connection.requeue_msg(msg)
                    yield None  # This will let other threads try to get the message

        def __find_ack(self, timeout=None):
            while(True):
                try:
                    msg = self.__connection.recv_msg(timeout=None)
                    if(msg.type == MSGType.ACKNOWLEDGE):  # There should always be only a single ack coming back
                        self._id = msg.request_id
                        break
                    else:
                        self.__connection.requeue_msg(msg)
                        time.sleep(0)  # some other thread probably wanted that message...
                except ZMQPairTimeout:
                    raise ZMQPairTimeout("Slave did not acknowledge request in {}s".format(timeout))


def _new_connection(conn):
    # Checks to see if connection with the same ID exists. If so, will close it and set the new ID. Otherwise just set
    # new ID
    conn_id = "{}:{}".format(conn.dest_ip, conn.port)
    try:
        connection = _get_connection(conn_id)
        connection.close()
    except RuntimeError:
        pass
    _pool.__dict__.setdefault("conplyent_pool", dict())[conn_id] = conn
    return conn_id


def _get_connection(conn_id):
    # Gets the connection from local pool.
    try:
        return getattr(_pool, "conplyent_pool")[conn_id]
    except (AttributeError, KeyError, IndexError):
        raise RuntimeError("Connection {} has not been set up".format(conn_id))


def _wrap_server_methods(cls, name):
    # Used to set new commands as attributes to the client connection
    def wrapper(self, *args, **kwargs):
        return self._command_server(name, *args, **kwargs)
    return wrapper



'''
:File: exceptions.py
:Author: Jayesh Joshi
:Email: jayeshjo1@utexas.edu
'''

import sys
import traceback


class ZMQPairTimeout(RuntimeError):
    '''
    Lower level timeout raised by ZMQPair when sending or receiving messages.
    Client/Server implementation should intercept this so users shouldn't need
    to worry about it.
    '''
    pass


class ConsoleExecTimeout(RuntimeError):
    '''
    Timeout raised by ConsoleExecutor if the executor did not respond in the
    amount of time specified by user.
    '''
    pass


class ClientTimeout(RuntimeError):
    '''
    Timeout raised by Client when the server did not repsond in the amount of
    time specified by user.
    '''
    pass


def exception_to_str():
    '''
    When python throws an error, users can catch the error and call this
    function to retrieve the entire call trace and the exception message in
    string format.

    :returns: String format detailing call stack and exception message.
    '''
    exc_type, exc_value, exc_traceback = sys.exc_info()
    return ''.join(traceback.format_exception(exc_type, exc_value, exc_traceback))


def min_exception_to_str():
    '''
    When python throws an error, users can catch the error and call this
    function to retrive only the exception message in string format.

    :returns: String format detailing the exception message.
    '''
    exc_type, exc_value, exc_traceback = sys.exc_info()
    return exc_value



from functools import wraps
import time


def timeout(_func=None, name="Timeout"):
    def timeout_decorator(func):
        @wraps(func)
        def timeout_wrapper(*args, **kwargs):
            if("timeout" in kwargs and kwargs["timeout"] is not None):
                start_time = time.perf_counter()
                for none in func(*args, **kwargs):
                    time_taken = time.perf_counter() - start_time
                    if(time_taken > kwargs["timeout"]):
                        raise kwargs["exception"]("{} did not complete in {}s".format(name, time_taken))
                    time.sleep(0.001)
            else:
                for none in func(*args, **kwargs):
                    time.sleep(0.001)
        return timeout_wrapper

    return timeout_decorator if(_func is None) else timeout_decorator(_func)



'''
:File: console.py
:Author: Jayesh Joshi
:Email: jayeshjo1@utexas.edu
'''

from subprocess import Popen, PIPE, STDOUT
from threading import Thread
from queue import Queue

from ._decorators import timeout
from .exceptions import ConsoleExecTimeout


class ConsoleExecutor():
    '''
    Simple wrapper around subprocess to provide a non-blocking read from
    stdout and stderr. This executor will start the subprocess using
    universal_newlines=True, shell=False, and start_new_session=True to provide
    the most responsive and proper executor. If the user specifies any kwargs
    that should be sent to Popen, then they can override these settings.

    This subprocess will start as the class is initialized and the class will
    start a background thread that continually reads the output from the
    executor. This way, if the user wants to read the output, we can simply
    check to see if the background thread has done anything by timing out this
    main thread and allowing the background thread to do its job. This also
    allows users to timeout any read requests if user just wants to check if
    there are any output.

    This class also provides other interactive means to communicate with the
    subprocess such as sending input and terminating.
    '''

    def __init__(self, cmd, **kwargs):
        self._cmd = cmd
        if(kwargs):
            self.__popen = Popen(cmd, stdout=PIPE, stderr=STDOUT, stdin=PIPE, **kwargs)
        else:
            self.__popen = Popen(cmd, stdout=PIPE, stderr=STDOUT, stdin=PIPE, universal_newlines=True, shell=False,
                                 start_new_session=True)
        self.__queue = Queue()
        self.__bg_worker = Thread(target=ConsoleExecutor.__file_reader, args=(self.__queue, self.__popen.stdout),
                                  daemon=True)
        self.__bg_worker.start()
        self._alive = True

    @property
    def cmd(self):
        '''
        Command linked to this executor.

        :getter: (int command that is being executed.
        '''
        return self._cmd

    @property
    def returncode(self):
        '''
        Exit code from the subprocess. Only set once the subprocess has exited.

        :getter: (int) any exit code from process
        '''
        return self.__popen.returncode

    @property
    def alive(self):
        '''
        Determines if the subprocess is still alive or not. The background
        thread used to read in any output from the subprocess will have exited
        once the subprocess has completed.

        :getter: (bool) True if subprocess is still alive. False if completed or
            exited otherways.
        '''
        return self.__bg_worker.is_alive()

    @property
    def empty(self):
        '''
        Determines if there is no more output left in the bg executor.

        :getter: (bool) True if no more output. False otherwise
        '''
        return self.__queue.empty()

    def read_output(self, timeout=None):
        '''
        Reads the next output from the subprocess. These must be sent by flushed
        stdout or stderr.

        By default, this method will poll forever until the subprocess has
        passed any output. Users can define timeout to wait only a specific
        amount of time for the next output.

        :param timeout: Amount of time in seconds to wait for the next output.
        :type timeout: int

        :returns: Output read from the subprocess. This value will be None if
            the subprocess has exited.

        :raises ConsoleExecTimeout: If user specifies a non-None/non-Negative
            timeout and subprocess has not responded in time.
        '''
        if(not(self.__queue.empty())):
            return self.__queue.get()
        while(True):
            if(self.alive):
                self.__poll_queue(timeout=timeout, exception=ConsoleExecTimeout)
                if(self.__queue.empty()):
                    if(not(self.alive)):
                        self.__popen.wait()
                        if(not(self.__queue.empty())):
                            return self.__queue.get()
                    return None
                else:
                    return self.__queue.get()  # should never halt here...
            else:
                self.__popen.wait()
                if(not(self.__queue.empty())):
                    return self.__queue.get()
                return None

    def send_input(self, value, new_line=True):
        '''
        Allows users to send input to the subprocess. This input will be flushed
        into the subprocess to ensure that the input will be read. To acheive
        this, this method will automatically add an extra new line if the user
        hasn't specified a new line. This automatic behavior can be disabled by
        optional user input new_line

        :param value: Message to send to the subprocess.
        :type value: str
        :param new_line: True if method should add a new line if missing. False
            to ignore this feature.
        :type new_line: bool
        '''
        if(self.alive):
            self.__popen.stdin.write(value + "\n" if value[:-1] != "\n" and new_line else value)
            self.__popen.stdin.flush()

    def kill(self):
        '''
        Terminates the subprocess and waits for it to exit gracefully. Currently
        this will not stop any child processes spawned by our subprocess.
        '''
        self.__popen.terminate()
        self.__popen.wait()

    def close(self):
        '''
        Closes off any FDs open by this class to properly clear any memory used
        by this subprocess. Terminates subprocess if alive.
        '''
        if(self.alive):
            self.kill()
        self.__popen.stdin.close()
        self.__popen.stdout.close()

    @timeout(name="Polling Subprocess")
    def __poll_queue(self, **kwargs):
        while(self.__queue.empty() and self.__bg_worker.is_alive()):
            yield None

    def __file_reader(queue, file):
        for line in iter(file.readline, b'' or ''):
            queue.put(line)
        file.close()



'''
:File: server.py
:Author: Jayesh Joshi
:Email: jayeshjo1@utexas.edu
'''

import os
from glob import glob
from functools import wraps
from shutil import rmtree
from threading import Thread
from zmq import Again, ZMQError
from queue import Queue
from pathlib import Path

from ._zmq_pair import ZMQPair
from ._msg import MSGType, MSG
from ._general import SynchronizedDict, logger, os_name
from .exceptions import min_exception_to_str, ConsoleExecTimeout
from .console import ConsoleExecutor


MSG_PORT = {}  # contains a dict of function names and calls
JOBS = SynchronizedDict()  # contains a dict of jobs and their
_msg_num = {}  # Ensures that MSG queue in client can be properly ordered

# Default exit codes
INVALID_PARAMETER = -1
SUCCESS = 0
ERROR = 1

# Default Exec queue params
INPUT = 0
KILL = 1

# Job indexes
QUEUE_IDX = 0
THREAD_IDX = 1


def start(port=9922):
    '''
    Begins the server on the current process. This is a looped call and only
    exists on KeyboardInterrupt  or if a connected client has requested to close
    the server.

    This is a wrapper aorund the ZMQ library using the PAIR method to
    communicate with the client. This binds the socket to localhost and port
    specified and will continually listen to the port for any activity.
    Initially, the client should begin by establishing the connection and
    performing a handshake to ensure that server is alive. This is done within
    the ZMQPair class and this method doesn't need to worry about it.

    Once the connection has been established, this method will continually
    listen for any requests from the client. Requests available are broken into
    three categories: heartbeat, sync, and command. Heartbeat, also handled
    within ZMQPair, simply responds back to the client to let the client know
    that the server is alive and able to communicate. Sync lets the client know
    which commands are registered and able for client to run (more on this
    later). Commands allows the client to access any registered method which
    the server can begin executing. When the server receives a command, it will
    immediately respond with an acknowledge to let the client know that the
    message has been received. Then, will continue executing the command. These
    commands can be foreground or background commands. Many commands can be
    performed in the foreground if they are not IO bound. For background tasks,
    server will update the client letting them know what job number has been
    assigned to this task so the client can receive messages and sort them out.
    There are also registered methods that can interact with the background
    tasks to affect their runtime.

    Finally, for the pre-defined registered method of "close_server", the server
    will update the client as per the method definition and then gracefully
    exit.

    :param port: Port to listen to in localhost.
    :type port: int

    :returns: SUCCESS if exiting gracefully.
    '''
    global _zmq_pair

    _zmq_pair = ZMQPair(port=port)
    _zmq_pair.bind()
    logger.info("SERVER START:: Begin listening on port {}".format(port))
    idx = 0

    while(True):
        msg = _zmq_pair.recv_msg()
        try:
            if(msg.type == MSGType.COMMAND):
                _zmq_pair.send_msg(MSG(MSGType.ACKNOWLEDGE, request_id=idx))
                MSG_PORT[msg.cmd_id](idx, *msg.args, **msg.kwargs)
                if(msg.cmd_id == "close_server"):
                    _zmq_pair.close()
                    return SUCCESS
                idx = (idx + 1) % 0xFFFFFFFF  # I really hope noone creates more than 4 mil jobs at once...
            elif(msg.type == MSGType.SYNC):
                _zmq_pair.send_msg(MSG(MSGType.SYNC, args=(list(MSG_PORT.keys()),)))
        except KeyboardInterrupt:
            _zmq_pair.close()  # make sure we close zmq gracefully before exit
            raise
        except Again:
            print("LOST")
            logger.info("Lost connection with host...")


def register_command(func):
    '''
    Decorator that registers the underlying function as a registered
    command that clients can execute remotely. This registration is used by the
    server to determine available commands.

    The wrapper will initially update the client and echo the request. Then,
    will execute the method and retrive the exit code. If there was an exception
    from the function, the wrapper will inform the client of the error and
    update the exit code to ERROR (1). Finally, the wrapper will report the
    completion of the function to the client reporting the exit code
    (INVALID_PARAMETER = -1, SUCCESS = 0, ERROR = 1).

    This module defines several commands that are available by default. Users
    can develop their own custom server commands using this decorator or even
    choose to clear default commands by clearing MSG_PORT if they so wish.

    By default this will provide the commands as accessible by their function
    name but am considering possibly allowing users to specify a name. (simple
    but need to create a few asserts to ensure it's a valid name <<).

    The functions that are registered under the server require that the first
    argument be idx. Idx refers to the job number that will be assigned once
    the function is called. This is important in updating the client to
    ensure that we are able to communicate to the client with the proper job
    number.

    From within the function, users can use calls to
    :meth:`server.update_client` to update the client with any output.

    .. note:: These functions will not include the idx in their docstring cuz
        it would be super redundant and don't help in their explanation
    '''
    function_name = func.__name__

    @wraps(func)
    def register_wrapper(idx, *args, **kwargs):
        request = "{} {} {}".format(function_name, args, kwargs)
        logger.info("Job {}: {}".format(idx, request))
        update_client(idx, request)

        try:
            exit_code = func(idx, *args, **kwargs)
        except Again:
            logger.info("Lost connection with host...")
            return
        except Exception:
            update_client(idx, min_exception_to_str())
            exit_code = 1
        _zmq_pair.send_msg(MSG(MSGType.COMPLETE, request_id=idx, exit_code=exit_code, msg_num=_msg_num[idx]))

    MSG_PORT[function_name] = register_wrapper
    return register_wrapper


def register_background_command(func):
    '''
    Decorator that registers the underlying function as a registered
    command that clients can execute remotely. This function will be executed on
    a background thread with a python Queue established between the main thread
    and the background thread for communication. These background commands are
    useful for functions that are IO bound or for those that will take longer
    than a few instructions (to prevent the server from blocking). Finally,
    these threads are ran as 'daemon' so exits alongside the server if the
    server prematurely exits.

    These background commands follow similar wrapper pattern to
    :meth:`server.register_command` in their exit code and error handling. These
    functions must return an exit code but do not need to worry about pre- and
    post- handling. Unlike the regular register_command decorator, this wrapper
    will add the command call to the JOBS dictionary as an active job which
    can be used for communcation using other commands.

    The function must have idx, and queue as their first two arguments.
    See register_command for details on idx, and how to update the client with
    output. Queue is used to communicate between the main server thread and
    these background threads if they require such communication. I.e. if we want
    to implement commands that require the thread's response. For default
    commands provided by conplyent, the queue and thread objects can be accessed
    through JOBS dict. This dict is accessible by providing job number as key
    and a tuple using QUEUE_IDX and THREAD_IDX to access the queue and thread
    objects respectively. The default queue recognition is by providing a
    dictionary that specifies "type" as either INPUT or KILL, and "value" if
    necessary. If developers chooses to create their own functions, they can
    redefine this queue as they see fit.

    .. note:: These functions will not include the idx or queue in their
        docstring cuz it would be super redundant and don't help in their
        explanation
    '''
    function_name = func.__name__

    def bg_server(idx, args, kwargs):
        request = "{} {} {}".format(function_name, args[1:], kwargs)
        logger.info("Job {} BG: {}".format(idx, request))
        update_client(idx, request)
        try:
            try:
                exit_code = func(idx, *args, **kwargs)
            except Again:
                logger.info("Lost connection with host...")
                JOBS[idx] = None
                return
            except Exception:
                update_client(idx, min_exception_to_str())
                exit_code = ERROR

            _zmq_pair.send_msg(MSG(MSGType.COMPLETE, request_id=idx, exit_code=exit_code, msg_num=_msg_num[idx]))
        except ZMQError:
            pass  # Client may have disconnected
        JOBS[idx] = None

    @wraps(func)
    def register_wrapper(idx, *args, **kwargs):
        io_pipe_r, io_pipe_w = os.pipe()
        my_queue = Queue()
        args = (my_queue,) + args
        t = Thread(target=bg_server, args=(idx, args, kwargs,), daemon=True)
        JOBS[idx] = (my_queue, t, args[1:], kwargs)
        t.start()

    MSG_PORT[function_name] = register_wrapper
    return register_wrapper


def update_client(idx, string):
    '''
    Sends the string to the client with the attached index to allow client to
    link messages to jobs. The string is sent as a pickled message object
    and doesn't have a upper limit on how long it could be. But y'know, better
    practice to put a limit on it. Sending huge messages over the network will
    eventually put a lower priority on the task by Windows (unless running in
    server environment.) This limit is pretty high though so for regular
    applications, users shouldn't worry.

    Note: this should never contain the exit code. Exit code will automatically
    sent by the wrapper.

    :param idx: Job number assigned by the server for this task.
    :type idx: int
    :param string: Message to send to the server as an update.
    :type string: str
    '''
    _msg_num.setdefault(idx, 0)
    _zmq_pair.send_msg(MSG(MSGType.DETAILS, request_id=idx, details=string, msg_num=_msg_num[idx]))
    _msg_num[idx] += 1


@register_command
def cd(idx, dest):
    '''
    Call os.chdir to be able to change current working directory of the server.
    If successful, will echo the new working directory back to the user and
    return SUCCESS exit code. Otherwise, will report to the user that server
    couldn't find the said directory and return INVALID_PARAM exit code.

    :param dest: Destination directory. Limitations for these are the same
        as os.chdir, in that it could be relative path or absolute. I.e.
        will support ".." to go backwards relatively or "/" to start at the
        root of this directory.
    :type dest: str

    :returns: SUCCESS if able to change directory. INVALID_PARAMETER if
        directory requested does not exist
    '''
    os.chdir(dest)
    update_client(idx, os.getcwd())
    return SUCCESS


@register_command
def ls(idx, src="."):
    '''
    Updates the client with the listing of the directory. This function
    currently only takes in a single parameter: src. If the user specifies the
    source to list, then ls will output the directory listing from there.
    Otherwise uses the current working directory. This display will specify
    whether the file found is a directory, file or a link and show the files
    in that order. If user provides a path and it doesn't exist as a directory,
    then will return INVALID_PARAM and let user know that directory does not
    exist. Otherwise, will return user with the listing of the directory and
    exit with SUCCESS.

    Future considerations -- expand it to show more details similar to BASH.

    :param src: Source path to list directory. Default cwd.
    :type src: str

    :returns: SUCCESS if able to list directory. INVALID_PARAMETER if directory
        requested does not exist
    '''
    if(not(os.path.isdir(src))):
        update_client("{}: No such directory".format(src))
        return INVALID_PARAMETER
    dir_path = os.getcwd() if src == "." else src
    name_list = "Listing of Directory {}:\n\n".format(dir_path)
    dir_list = []
    link_list = []
    file_list = []
    for file in glob("{}/*".format(dir_path)):
        file = os.path.basename(file)
        if(os.path.isdir(file)):
            dir_list.append(file)
        elif(os.path.islink(file)):
            link_list.append(file)
        else:
            file_list.append(file)
    name_list += "\n".join(["d {}".format(i) for i in dir_list]) + "".join(["\nf {}".format(i) for i in file_list]) +\
        "".join(["\nl {}".format(i) for i in link_list])
    update_client(idx, name_list)
    return SUCCESS


@register_command
def cwd(idx):
    '''
    Updates the client with the current working directory of the server.

    :returns: SUCCESS -- should never fail...
    '''
    update_client(idx, os.getcwd())
    return SUCCESS


@register_command
def mkdir(idx, path, mode=None):
    '''
    Creates a new empty directory under the path provided by the user. If the
    path exists, either as a file or a directory, then update the client that
    the path already exists and return INVALID_PARAMETER. Otherwise, create a
    directory and update the user that the directory has been created.

    Users may also pass in mode as either an argument or kwarg to pass into
    os.mkdir to change the mode of the file created. If the mode passed in is
    rejected by os.mkdir, server will report to the client of the error.

    :param path: Path of new directory to be created.
    :type path: str

    :Optional:

    :param mode: Mode of the directories to be created.
    :type mode: int

    :returns: SUCCESS if able to create a directory. INVALID_PARAMETER if there
        exists something in the path.
    '''
    if(not(os.path.exists(path))):
        if(mode is None):
            os.mkdir(path)
        else:
            os.mkdir(path, mode)
        update_client(idx, "Created directory: {}".format(path))
        return SUCCESS
    else:
        update_client(idx, "Directory or File {} already exists".format(path))
        return INVALID_PARAMETER


@register_command
def mkdirs(idx, path, exist_ok=False):
    '''
    Goes through the whole path, creating directories if they don't exist. By
    default, this will throw an error if the end directory already exists. Users
    can pass in exist_ok to prevent this error.

    :param path: Path of new directory to be created
    :type path: str

    :Optional:

    :param exist_ok: set to True if not to throw error if exists
    :type exist_ok: bool

    :returns: SUCCESS
    '''
    Path(path).mkdir(parents=True, exist_ok=exist_ok)
    update_client(idx, "Created directory: {}".format(path))
    return SUCCESS


@register_command
def rm(idx, path, recursive=False):
    '''
    Removes files or directories or links within the path specified. Initially,
    will check to see if anything exists in the path. If it doesn't then report
    to the client that there is nothing to be removed in the path provided and
    return INVALID_PARAMETER. Otherwise, if the path points to a file or link,
    then remove it, update the client and return SUCCESS. If the path points to
    a directory, then check if the directory has any contents within it. If it
    doesn't then we can safely remove it and update the client + return SUCCESS.
    Otherwise, remove the directory and all of its content if user has passed in
    recursive, update the cleint + return SUCCESS. If the client didn't specify
    recursive (or noted to not do recursive), then let the client know that the
    directory is not empty and return INVALID_PARAMETER.

    :param path: Path of file/directory to remove.
    :type path: str

    :Optional:

    :param recursive: True if we should remove all contents within a directory.
        False to exit if directory contains anything.
    :type recursive: bool

    :returns: SUCCESS if able to remove successfully. INVALID_PARAMETER if no
        file/directory exists in the path or recursive is False and path
        contains a non-empty directory.
    '''
    if(os.path.exists(path)):
        if((os.path.isdir(path) and recursive) or (os.path.isdir(path) and not(os.listdir(path)))):
            rmtree(path)
            update_client(idx, "Removed directory {}".format(path))
            return SUCCESS
        elif(not(os.path.isdir(path))):
            os.remove(path)
            update_client(idx, "Removed file {}".format(path))
            return SUCCESS
        else:
            update_client(idx, "Non-Empty Directory... Pass recursive to recursivly remove directory")
            return INVALID_PARAMETER
    else:
        update_client(idx, "Directory or File {} doesn't exist?".format(path))
        return INVALID_PARAMETER


@register_command
def rdfile(idx, path, limit=1024):
    '''
    Opens file in the path that the user provided and reads it contents. First,
    this function will echo "Contents of {path provided by user} and updates
    the user for each KB of data in the file contained. This data is sent as
    binary and left to the client to decode as necessary. Once complete, will
    will return SUCCESS. If the path doesn't point to a file, then will inform
    the client and return INVALID_PARAMETER.

    :param path: Path to file to read.
    :type path: str

    :Optional:

    :param limit: Number in bytes to send per transaction. This is default 1 KB.
        Higher numbers may slow down network and increase the chances of packet
        loss.
    :type limit: int

    :returns: SUCCESS if able to read the file. INVALID_PARAM if file doesn't
        exist.
    '''
    if(os.path.exists(path) and not(os.path.isdir(path))):
        with open(path, 'rb') as file:
            update_client(idx, "Contents of {}".format(path))
            while(True):
                chunk = file.read(limit)
                if(chunk):
                    update_client(idx, chunk)
                else:
                    break
        return SUCCESS
    else:
        update_client(idx, "Unknown file: {}".format(path))
        return INVALID_PARAMETER


@register_command
def wrfile(idx, path, data, append=False):
    '''
    Opens file in path for write and flushes the data provided. This data must
    be in binary and the file must exist in path. By default this will create a
    new file decriptor under the name provided and flush the data. This would
    mean that if the file exists, then this call would delete the old file and
    add the data provided. Users can bypass this by providing append=True to
    append to the end of file instead.

    This call will check to see if the path exists as a directory, in which case
    this will notify the client and return INVALID_PARAMETER. Otherwise, will
    begin to write all of the data, let the user know of completion and return
    SUCCESS. This may also throw an error if the input provided by the user
    is not writable.

    :param path: Path to file to write
    :type path: str
    :param data: Data in bytes to write to file
    :type data: bytes

    :Optional:

    :param append: True to append to end of file if file exists. False default.
    :type append: bool

    :returns: SUCCESS if able to write into file. INVALID_PARAMETER if path
        exists as a directory.
    '''
    if(os.path.exists(path) and os.path.isdir(path)):
        update_client(idx, "wrfile: Path exists as a directory.")
        return INVALID_PARAMETER
    else:
        with open(path, "ab" if append else "wb") as file:
            if(append):
                file.seek(0, 2)
            file.write(data)
        update_client(idx, "Finished writing data to {}".format(path))
        return SUCCESS


@register_command
def touch(idx, path):
    '''
    Similar in function to Linux's touch. If the user passes in a path that
    points to a directory, then will notify the client and return
    INVALID_PARAMETER/. Otherwise, firstly it checks to see if the path points
    to a file or not. If it doesn't point to a file, this function will create
    an empty file. If the file exists, then this function will update the
    modification time on the file without changing its content. In both cases,
    will report to the client of completion and return SUCCESS.

    :param path: Path to file to touch.
    :type path: str

    :returns: SUCCESS if able to touch file. INVALID_PARAMETER if path exists as
        a directory.
    '''
    if(os.path.exists(path) and os.path.isdir(path)):
        update_client(idx, "wrfile: Path exists as a directory.")
        return INVALID_PARAMETER
    else:
        file = open(path, "ab+" if os.path.exists(path) else "wb+")
        file.write(b" ")
        file.flush()
        file.seek(-1, 2)
        file.truncate()
        file.close()
        update_client(idx, "Touched {}".format(path))
        return SUCCESS


@register_command
def jobs(idx):
    '''
    Checks to see all the JOBS that are still running in the server as
    background taks and notifies the client of them.

    :returns: SUCCESS
    '''
    output = ""
    for key, value in JOBS.items():
        if(value is not None):
            output += "Running {}: {}\n".format(key, value[2:])
    if(output == ""):
        output = "No jobs running."
    update_client(idx, output)
    return SUCCESS


@register_background_command
def exec(idx, queue, cmd):
    '''
    Runs as a background command. This function is used to run commands on the
    system by executing commands directly. Exec begins by creating a new
    :class:`console.ConsoleExecutor` and begins listening to any output. Every
    10 ms or every time it receives an output, exec will check to see if there
    are any traffic on the queue linked to the main server. Currently, the
    server may pass in any input to send to the subprocess running the external
    command or for the command to kill the subprocess.

    For any output read by exec, it will notify the client. Finally, when the
    subprocess exits, will notify the client for the completion.

    :param cmd: Command in string to run on the system. These are unique to each
        OS so users should know beforehand what the usecase is.
    :type cmd: str

    :returns: SUCCESS
    '''
    m_executor = ConsoleExecutor(cmd)
    while(m_executor.alive or not(m_executor.empty)):
        try:
            line = m_executor.read_output(0.01)
            if(line is not None):
                update_client(idx, line)
        except ConsoleExecTimeout:
            pass
        if(not(queue.empty())):
            command = queue.get()
            if(command["type"] == INPUT):
                m_executor.send_input(command["value"])
            elif(command["type"] == KILL):
                m_executor.close()
                break
    m_executor.close()
    return m_executor.returncode


@register_command
def send_input(idx, job_num, value):
    '''
    For the jobs that follow the default queue pattern, this method sends a
    INPUT command to the background job with the associated value. The
    background threads will then read this command and send them to the
    corresponding process. I.e. for exec, since exec is listening for output
    from the subprocess, exec bg thread is also listening for any traffic in
    the queue. If the job doesn't exist or has already completed, will notify
    the user of such and return INVALID_PARAMETER. Otherwise, push the request
    into the queue and return success.

    :param job_num: ID of the job to pass this message to.
    :type job_num: int
    :param value: Send input as string to the job.
    :type value: str

    :returns: SUCCESS if able to send input to job. INVALID_PARAMETER if job
        doesn't exist or has already exited.
    '''
    if(job_num in JOBS and JOBS[job_num] is not None):
        JOBS[job_num][QUEUE_IDX].put({"type": INPUT, "value": value})
        update_client(idx, "Sent value to job {}".format(job_num))
        return SUCCESS
    else:
        update_client(idx, "Job does not exist or has already completed...")
        return INVALID_PARAMETER


@register_command
def kill(idx, job_num):
    '''
    For the jobs that follow the default queue pattern, this method sends a
    KILL command to the background job with the associated value. The
    background threads will then read this command and terminate the
    corresponding process. I.e. for exec, since exec is listening for output
    from the subprocess, exec bg thread is also listening for any traffic in
    the queue.

    :param job_num: ID of the job to pass this message to.
    :type job_num: int

    :returns: SUCCESS if able to send kill command to job. INVALID_PARAMETER if
        job doesn't exist or has already exited.
    '''
    if(job_num in JOBS and JOBS[job_num] is not None):
        JOBS[job_num][QUEUE_IDX].put({"type": KILL})
        update_client(idx, "Sent kill command to job {}".format(job_num))
    else:
        update_client(idx, "Job does not exist or has already completed...")
        return INVALID_PARAMETER


@register_command
def close_server(idx):
    '''
    Kills any jobs that are running in the background and awaits for them to
    respond. Updates the client with all of the jobs that have been killed.

    Finally, updates the client of the incoming closure and returns SUCCESS. The
    main loop of the server will recognize that close_server has been called and
    exit.

    :returns: SUCCESS
    '''
    for key, job_tuple in JOBS.items():
        if(job_tuple is not None):
            job_tuple[QUEUE_IDX].put({"type": KILL})
            update_client(idx, "Killing job {}".format(key))
            job_tuple[THREAD_IDX].join()

    update_client(idx, "Closing Server")
    return SUCCESS


@register_command
def os_info(idx):
    '''
    Can be used to determine the type of OS running on the server. Updates the
    client with "windows", "linux", or "mac".

    :returns: SUCCESS
    '''
    update_client(idx, os_name())
    return SUCCESS


@register_command
def user_name(idx):
    '''
    Used to determine the name of user who's logged in, mainly for local directory
    naming.

    :returns: SUCCESS
    '''
    update_client(idx, os.getlogin())
    return SUCCESS


@register_command
def reboot(idx):
    '''
    Issues a reboot command to the server based on which OS this server is
    running on. For all of the cases, we simply notify the user of the
    inevitable shutdown of the system.

    :returns: SUCCESS

    .. note:: This job may not respond with COMPLETE, depending on how fast the
        OS reboots the system...
    '''
    my_os = os_name()
    update_client(idx, "Requesting os to reboot")
    if(my_os == "windows"):
        os.system("shutdown -t 0 -r -f")
    elif(my_os == "linux"):
        os.system("/sbin/reboot now")
    elif(my_os == "mac"):
        raise NotImplementedError("Macs not supported at the moment")
    return SUCCESS


@register_command
def shutdown(idx):
    '''
    Issues a shutdown command to the server based on which OS this server is
    running on. For all of the cases, we simply notify the user of the
    inevitable shutdown of the system.

    :returns: SUCCESS

    .. note:: This job may not respond with COMPLETE, depending on how fast the
        OS shutdowns the system...
    '''
    my_os = os_name()
    update_client(idx, "Requesting os to shutdown")
    if(my_os == "windows"):
        os.system("shutdown -t 0 -s -f")
    elif(my_os == "linux"):
        os.system("/sbin/shutdown now")
    elif(my_os == "mac"):
        raise NotImplementedError("Macs not supported at the moment")
    return SUCCESS



import socket
import os
import logging
from threading import Lock

logger = logging.getLogger("conplyent")


class SynchronizedDict(dict):
    def __init__(self, *args, **kwargs):
        self.__lock = Lock()
        super(SynchronizedDict, self).__init__(*args, **kwargs)

    def __getitem__(self, key):
        self.__lock.acquire()
        try:
            return super(SynchronizedDict, self).__getitem__(key)
        finally:
            self.__lock.release()

    def __setitem__(self, key, value):
        self.__lock.acquire()
        try:
            super(SynchronizedDict, self).__setitem__(key, value)
        finally:
            self.__lock.release()


def ipv4():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(("8.8.8.8", 80))
        host_name = s.getsockname()[0]
        s.close()
        return host_name
    except KeyboardInterrupt:
        raise
    except Exception:
        return None


def os_name():
    return {"nt": "windows", "posix": "linux", "mac": "mac"}[os.name]



#!/usr/bin/env python

'''
:File: cli.py
:Author: Jayesh Joshi
:Email: jayeshjo1@utexas.edu

Console script used to install conplyent as a startup program on both Windows
and Linux, and provides means to run both the client and the server.
'''

import os
import ast
import re
import logging
from subprocess import run, CREATE_NEW_CONSOLE

import click
import conplyent


def _install_windows(port):
    print("Detected Windows OS")
    print("Installing conplyent server listening to port # {}".format(port))
    user = os.getlogin()
    startup = "{}\\..\\Users\\{}\\AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs\\Startup".format(
        os.environ.get("windir"), user)  # This works for Windows 10... not sure about Windows 7-
    print("Assumming startup folder is in {}".format(startup))
    file_name = "{}\\conplyent_{}.bat".format(startup, port)
    with open(file_name, "w") as file:
        file.write("if not DEFINED IS_MINIMIZED set IS_MINIMIZED=1 && start \"\" /min \"%~dpnx0\" %* && exit\n"
                   "    conplyent start-server --port {}\n".format(port) +
                   "exit")
    print("Created new file {}".format(file_name))

    print("Starting new console to start off the server")
    run("{}\\conplyent_{}.bat".format(startup, port), creationflags=CREATE_NEW_CONSOLE)


def _install_linux(port):
    import lsb_release
    release = lsb_release.get_lsb_information()["RELEASE"]
    print("Detected Linux OS {}".format(release))
    if(not(release[:2] == "16" or release[:2] == "18")):
        print("Linux OS not currently supported for installation")
        return
    print("Installing conplyent server listening to port # {}".format(port))

    if(not(os.path.isdir("/usr/bin/conplyent"))):
        os.mkdir("/usr/bin/conplyent")

    with open("/usr/bin/conplyent/conplyent_{}.sh".format(port), "w") as file:
        file.write("#!/bin/sh\nconplyent start-server --port {}".format(port))

    os.system("chmod +x /usr/bin/conplyent/conplyent_{}.sh".format(port))

    with open("/etc/systemd/system/conplyent_{}.service".format(port), "w") as file:
        file.write("[Unit]\nDescription=conplyent service\nWants=network-online.target\n"
                   "After=network-online.target network.target\n\n")
        file.write("[Service]\nExecStart=/usr/bin/conplyent/conplyent_{}.sh\n\n".format(port))
        file.write("[Install]\nWantedBy=default.target")

    os.system("chmod 664 /etc/systemd/system/conplyent_{}.service".format(port))
    os.system("systemctl enable conplyent_{}.service".format(port))

    print("Done")


def _parse_args(arg_list):
    args = []
    kwargs = {}

    for arguments in arg_list[1:]:
        if(arguments):
            if("=" in arguments):
                key, value = arguments.split("=")
                try:
                    kwargs[key] = ast.literal_eval(value)
                except (ValueError, SyntaxError):
                    kwargs[key] = value
            else:
                try:
                    args.append(ast.literal_eval(arguments))
                except (ValueError, SyntaxError):
                    args.append(arguments)

    return args, kwargs


@click.group()
def cli():
    pass


@cli.command(help="Installs the server to startup on each boot")
@click.option("-p", "--port", help="Startup server will run on specified port", default=9922, type=int)
def install(port):
    '''
    Installs conplyent server for each platform. This currently supports Windows
    10 and Ubuntu 16.04+. On windows, this simply adds it to the startup local
    location. On ubuntu, this registers the server as a systemd service.
    '''
    if(os.name == "nt"):
        _install_windows(port)
    elif(os.name == "posix"):
        _install_linux(port)
    else:
        raise NotImplementedError("Unknown OS... unsupported by conplyent at the moment")


@cli.command(help="Uninstalls the server and prevents it from starting up")
@click.option("-p", "--port", help="Define port to uninstall", default=9922, type=int)
def uninstall(port):
    raise NotImplementedError("To be implemented")


@cli.command(name="start-server", help="Runs the server and starts listening on port")
@click.option("-p", "--port", help="Starts server on specified port", default=9922, type=int)
@click.option("--quiet", help="Sets the logging to quiet", default=False, is_flag=True)
@click.option("--debug", help="Sets the logging to debug (quiet must be false)", default=False, is_flag=True)
def start_server(port, quiet, debug):
    '''
    Starts the server on localhost on the provided port. Users can run this on
    quiet mode, basic info mode, or on debug mode.
    '''
    if(not(quiet)):
        logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)

    conplyent.server.start(port)


@cli.command(name="start-client", help="Run client to talk to server")
@click.option("-h", "--hostname", help="Host name of server to connect to", required=True, type=str)
@click.option("-p", "--port", help="Starts server on specified port", default=9922, type=int)
@click.option("-t", "--timeout", help="Timeout waiting for server to connect", default=None, type=int)
def start_client(hostname, port, timeout):
    '''
    Starts interactive client mode. Only keyword connected to this interactive
    mode is "commands" which will print out all available server commands.
    '''
    conn = conplyent.client.add(hostname, port)
    conn.connect(timeout=timeout)
    server_methods = conn.server_methods()

    while(True):
        response = input("Enter command: ")
        arg_list = re.split(r"\[(.*)\]|\"(.*)\"|\'(.*)\'| ", response)
        if(arg_list and (arg_list[0] == "commands")):
            print("Server commands:\n{}".format("\n".join(server_methods)))
        elif(not(arg_list) or not(arg_list[0] in server_methods)):
            print("Unknown command")
        else:
            args, kwargs = _parse_args(arg_list)
            getattr(conn, arg_list[0])(*args, **kwargs, complete=True, echo_response=True)


if(__name__ == '__main__'):
    cli()



'''
:File: _zmq_pair.py
:Author: Jayesh Joshi
:Email: jayeshjo1@utexas.edu

Wrapper around the ZMQ class used by client and server to communicate.

Built on the principle of ISRs -- IO requests should be handled in the
background when the main thread is not busy. Main thread can then poll the
background thread to check to see if there are any updates. This allows the main
thread to run foreground activities and check the connection whenever desired
for any updates.
'''

import zmq
import time
from threading import Thread, local
from queue import Queue

from ._msg import MSG, MSGType
from ._decorators import timeout
from ._general import logger
from .exceptions import ZMQPairTimeout


_ctx = local()


class ZMQPair(object):
    def __init__(self, dest_ip=None, port=9922):
        self._context = _get_context()
        self._socket = self._context.socket(zmq.PAIR)
        self._socket.setsockopt(zmq.LINGER, 0)
        self._dest_ip = dest_ip
        self._port = port
        self._connected = False
        self.__queue = Queue()
        self.__msg_backlog = {}
        self.__bg_worker = Thread(target=ZMQPair.__bg_receiver, args=(self.__queue, self._socket), daemon=True)
        self.__bg_worker.start()

    @property
    def context(self):
        return self._context

    @property
    def socket(self):
        return self._socket

    @property
    def dest_ip(self):
        return self._dest_ip

    @property
    def port(self):
        return self._port

    @property
    def connected(self):
        return self._connected

    def close(self):
        if(self._connected and self._dest_ip):
            self.disconnect()
        self._socket.close()

    def bind(self):
        self._connected = self.__heartbeat(0.1)
        if(not(self._connected)):
            self._socket.bind("tcp://*:{}".format(self._port))
            self._connected = True

    def unbind(self):
        if(self._connected):
            self._socket.unbind()
            self._connected = False

    def connect(self):
        self._connected = self.__heartbeat(0.1)
        if(not(self._connected)):
            self._socket.connect("tcp://{}:{}".format(self._dest_ip, self._port))
            self._connected = True
            logger.debug("Main:: Connected: {}:{}".format(self._dest_ip, self._port))

    def disconnect(self):
        self._socket.disconnect("tcp://{}:{}".format(self._dest_ip, self._port))
        self._connected = False
        logger.debug("Main:: Disconnect: {}:{}".format(self._dest_ip, self._port))

    def send_msg(self, msg, timeout=None):
        assert type(msg) == MSG, "ZMQPair only communicates through MSG class"
        if(self._connected):
            logger.debug("Main:: Sending Message {}".format(str(msg)))
            self.__send_process(msg, timeout=timeout, exception=ZMQPairTimeout)
            return True
        else:
            self._connected = False
            return False

    def recv_msg(self, timeout=None, msg_id=None):
        if(msg_id is not None and msg_id in self.__msg_backlog and self.__msg_backlog[msg_id]):
            return self.__msg_backlog[msg_id].pop(0)
        self.__check_mail(timeout=timeout, exception=ZMQPairTimeout)
        mail = self.__queue.get(timeout=10)
        logger.debug("Main:: Retrieved Message: {}".format(str(mail)))
        return mail

    def requeue_msg(self, msg):
        if(not(msg.type == MSGType.ACKNOWLEDGE) and msg.has_request_id()):
            self.__msg_backlog.setdefault(msg.request_id, list()).append(msg)
            logger.debug("Putting MSG to backlog {}".format(msg))
        else:
            self.__queue.put(msg)

    def pulse(self, timeout=None):
        return self.__heartbeat(timeout=timeout)

    @timeout(name="Trasmitter")
    def __send_process(self, msg, **kwargs):
        tracker = self._socket.send_pyobj(msg, flags=zmq.NOBLOCK, track=True, copy=False)
        while(not(tracker.done)):
            yield None

    @timeout(name="Receiver")
    def __check_mail(self, **kwargs):
        if(self.__bg_worker.is_alive()):
            while(self.__queue.empty()):
                yield None
        else:
            raise RuntimeError("Recv thread died?")

    def __heartbeat(self, timeout):
        if(self._connected):
            self.send_msg(MSG(MSGType.HEARTBEAT, request=True), timeout=timeout)
            try:
                while(True):
                    msg = self.recv_msg(timeout=timeout)
                    if(msg.type == MSGType.HEARTBEAT):
                        return True
                    else:
                        self.__queue.put(msg)
            except ZMQPairTimeout:
                return False
        else:
            return False

    def __bg_receiver(queue, socket):
        logger.debug("Thread:: Starting")
        try:
            while(True):
                while(not(socket.poll(timeout=1, flags=zmq.POLLIN))):
                    time.sleep(0)
                msg = socket.recv_pyobj()
                if(msg.type == MSGType.ACKNOWLEDGE and msg.request_id == 0):  # if server reset, clear out older msgs
                    queue_return = []
                    while(not(queue.empty())):
                        msg = queue.get()
                        if(msg.type == MSGType.HEARTBEAT or msg.type == MSGType.SYNC):
                            queue_return.append(msg)
                    while(queue_return):
                        queue.put(queue_return.pop(0))
                logger.debug("Thread:: {}".format(str(msg)))
                if(msg.request and msg.type == MSGType.HEARTBEAT):
                    socket.send_pyobj(MSG(MSGType.HEARTBEAT, request=False))
                else:
                    queue.put(msg)
        except zmq.ZMQError:
            pass


def _get_context():
    try:
        return getattr(_ctx, "zmq_context")
    except (AttributeError, IndexError):
        _ctx.__dict__["zmq_context"] = zmq.Context()
        return _ctx.zmq_context


def _close_context():
    try:
        getattr(_ctx, "zmq_context").term()
    except (AttributeError, IndexError):
        pass  # No need to close context?



from setuptools import setup


setup(
    name="conplyent",
    version="0.1.1",
    license="MIT",
    author="Jayesh Joshi",
    author_email="jayeshjo1@utexas.edu",
    url="https://github.com/joshijayesh/conplyent/",
    description="Local and Remote Console executor",
    packages=["conplyent", "conplyent_scripts"],
    install_requires=[
        "pyzmq>=17.1.2",
        "click>=7.0"],
    entry_points={
        "console_scripts": [
            "conplyent=conplyent_scripts.cli:cli"
        ],
    },
    python_requires=">=3.4.3",
)



from enum import Enum


class MSGType(Enum):
    ACKNOWLEDGE = 0,
    HEARTBEAT = 1,
    COMMAND = 2,
    DETAILS = 3,
    COMPLETE = 4,
    SYNC = 5


class MSG(object):
    def __init__(self, type, request=True, **kwargs):
        self._type = type
        self._request = request
        self._kwargs = kwargs

    def __str__(self):
        string = "<MSG Type: {}> <Request?: {}>".format(self._type, self._request)
        if(self._kwargs):
            string += " <Kwargs: {}>".format(self._kwargs)
        return string

    @property
    def type(self):
        return self._type

    @property
    def request(self):
        return self._request

    @property
    def cmd_id(self):
        return self._kwargs["cmd_id"]

    @property
    def args(self):
        return self._kwargs["args"]

    @property
    def kwargs(self):
        return self._kwargs["keywargs"]

    @property
    def details(self):
        return self._kwargs["details"]

    @property
    def request_id(self):
        return self._kwargs["request_id"]

    @property
    def exit_code(self):
        return self._kwargs["exit_code"]

    @property
    def msg_num(self):
        return self._kwargs["msg_num"]

    def has_request_id(self):
        return "request_id" in self._kwargs
